#include <EXTERN.h>
#include <perl.h>
#include "amc.h"
#include "progeny.h"
#define and &&
#define or ||

static PerlInterpreter *my_perl;
void runasub (char* subname);

unsigned short baseindex, nextindex;
AM_SUPRA **lattice, *supralist, *HETERO;

int main(int argc, char **argv) {
  static char *dummyargv[] = {"", "amc.pl"};
  short num_test_items, repeat_count, maxvar;
  HETERO = (AM_SUPRA *) malloc(sizeof(AM_SUPRA));
  supralist = (AM_SUPRA *) malloc(sizeof(AM_SUPRA));
  supralist->index = 0;
  supralist->outcome = 0;
  supralist->data = (AM_CONTEXT *) malloc(sizeof(AM_CONTEXT));
  supralist->data[0] = 0;
  supralist->count = 0;
  supralist->next = supralist;

  my_perl = perl_alloc();
  perl_construct(my_perl);
  perl_parse(my_perl, NULL, 2, dummyargv, (char **) NULL);
  perl_run(my_perl);

  runasub("setup");
  num_test_items = av_len(get_av("testItems", FALSE)) + 1;
  repeat_count = SvIV(get_sv("repeat", FALSE));
  maxvar = SvIV(get_sv("maxvar", FALSE));
  printf("Attempting to allocate memory for %i variables...", maxvar);
  lattice = (AM_SUPRA **) calloc(1 << maxvar, sizeof(AM_SUPRA *));
  if (lattice == NULL) {
    printf("failed.\n");
    goto FREE;
  } else {
    printf("succeeded.\n");
  }

  for (; num_test_items; --num_test_items) {
    int activeVar;
    AM_CONTEXT activeContexts;
    short rc = repeat_count;

    runasub("parseTestItem");
    activeVar = SvIV(get_sv("activeVar", FALSE));
    activeContexts = SvIV(get_sv("activeContexts", FALSE));

    for(; rc; --rc) {
      AM_CONTEXT numsubcontexts, i, d;
      AM_CONTEXT *subcontext;
      AM_OUTCOME *suboutcome;
      runasub("beginTestItem");
      numsubcontexts = SvIV(get_sv("numsubcontexts", FALSE));
      if (!numsubcontexts) {
	runasub("endTestItem");
	continue;
      }
      for (i = 0; i < activeContexts; ++i) lattice[i] = supralist;
      subcontext = (AM_CONTEXT *) SvPV_nolen(get_sv("subcontexts", FALSE));
      subcontext += numsubcontexts;
      suboutcome = (AM_OUTCOME *) SvPV_nolen(get_sv("suboutcomes", FALSE));
      suboutcome += numsubcontexts;
      nextindex = 1;

      while (numsubcontexts) {
	--numsubcontexts;
	--subcontext;
	--suboutcome;
	baseindex = nextindex;

	if (lattice[*subcontext] == HETERO) continue;
	add(*subcontext, *subcontext, *suboutcome);
	ancestor(*subcontext, activeVar);
	while (d = descendant()) {
	  if (lattice[d] == HETERO) continue;
	  add(d, *subcontext, *suboutcome);
	}
	cleansupra();
      }
      countsupra();
      runasub("endTestItem");
    }
  }

  runasub("closeshop");

 FREE:
  PL_perl_destruct_level = 0;
  perl_destruct(my_perl);
  perl_free(my_perl);

  free(supralist->data);
  free(supralist);
  free(lattice);
  free(HETERO);
}

void runasub (char* subname) {
  static char** dummyargs = (char **) NULL;
  dSP;
  call_argv(subname, G_DISCARD | G_NOARGS, dummyargs);
}

void add(AM_CONTEXT supracontext, AM_CONTEXT subcontext, AM_OUTCOME outcome) {
  AM_SUPRA *p, *c;

  p = lattice[supracontext];
  if (p->count) --(p->count);
  if (p->next->index < baseindex) {
    c = (AM_SUPRA *) malloc(sizeof(AM_SUPRA));
    c->index = nextindex++;
    c->count = 0;
    c->next = p->next;
    p->next = c;

    if((outcome and outcome == p->outcome) or !p->index) {
      c->outcome = outcome;
      ++(c->count);
      c->data = calloc(p->data[0] + 2, sizeof(AM_CONTEXT));
      memcpy(c->data, p->data, (p->data[0] + 1) * sizeof(AM_CONTEXT));
      c->data[++(c->data[0])] = subcontext;
      lattice[supracontext] = c;
    } else {
      c->outcome = 0;
      c->data = NULL;
      lattice[supracontext] = HETERO;
    }
  } else {
    if (p->next->data != NULL) {
      ++((lattice[supracontext] = p->next)->count);
    } else {
      lattice[supracontext] = HETERO;
    }
  }
}

void cleansupra(void) {
  AM_SUPRA *p, *t;

  for (p = supralist; p->next != supralist;) {
    if (p->next->count == 0) {
      p->next = (t = p->next)->next;
      free(t->data);
      free(t);
    } else {
      p = p->next;
    }
  }
}

void countsupra(void) {
  AM_SUPRA *p, *t;

  for (p = supralist->next; p != supralist; ) {
    dSP;
    ENTER;
    SAVETMPS;
    PUSHMARK(SP);
    XPUSHs(sv_2mortal(newSViv(p->count)));
    XPUSHs(sv_2mortal(newSVpv((char *)(p->data + 1), p->data[0] * sizeof(AM_CONTEXT))));
    PUTBACK;
    call_pv("count", G_DISCARD);
    SPAGAIN;
    PUTBACK;
    FREETMPS;
    LEAVE;

    free(p->data);
    t = p;
    p = p->next;
    free(t);
  }
  supralist->next = supralist;
}
