############################################################
## a couple global variables

my @list;
my $tally;


############################################################
sub setup {

  srand(time() ^ ($$ + ($$ << 15)));

## set input record separator for slurping

  undef $/;

## config file

  open (CONFIG, "AM.config");
  $_ = <CONFIG>;
  close CONFIG;

  ($name) = /Project Name.*:\s*(.*)/;
  $datafile = "$name/data";
  $outcomefile = "$name/outcome";
  $testfile = "$name/test";
  $resultsfile = "$name/amcresults";

  ($excNull) = /null.*:\s*(.*)/;
  $excNull = 0 unless $excNull eq 'exclude';

  ($excGiven) = /given.*:\s*(.*)/;
  $excGiven = 0 unless $excGiven eq 'exclude';

  ($probability) = /probability.*:\s*(.*)/;
  ($repeat) = /repeat.*:\s*(.*)/;

  ($linearOrSq) = /linear or squared.*:\s*(.*)/;
  $linearOrSq eq 'linear' ? ($$tally = 0) : ($tally = \$#list);

  ($outputToFile) = /output to file.*:\s*(.*)/;
  if ($outputToFile eq 'yes') {
    open (PFLE, ">$resultsfile");
    select PFLE;
  }

  ($specifyFreq) = /specify frequency.*:\s*(.*)/;
  $specifyFreq = 0 if $specifyFreq eq 'no';

  ($usescommas) = /format with commas.*:\s*(.*)/;
  if ($usescommas eq 'yes') {
    $bigsep = '\s*,\s*';           # assume comma-separated
    $smallsep = '\s+';             # and space-separated
  } else {
    $bigsep = '\s+';               # assume space-separated
    $smallsep = '';                # and not separated
  }

## reset input record separator

  $/ = "\n";

## outcome file

  open(OUTCOME, "$outcomefile");
  @outcomelist = @ocl = (undef);
  my $outcomecounter = 0;
  while(<OUTCOME>) {
    chomp;
    my($oc, $outcome) = split;
    $outcometonum{$oc} = ++$outcomecounter;
    push @outcomelist, $outcome;
    push @ocl, $oc;
  }
  close OUTCOME;

## data file

  open(DATA, "$datafile");
  while(<DATA>) {
    chomp;
    my($outcome, $data, $spec) = split /$bigsep/, $_, 3;

    push @outcome, $outcometonum{$outcome};
    push @spec, $spec;
    push @data, [split /$smallsep/, $data];

    last if $specifyFreq and ($specifyFreq == @data);

  }
  close DATA;

## test file

  open(TEST, "$testfile");
  @testItems = <TEST>;
  close TEST;
  chomp(@testItems);
  (undef, $item) = split /$bigsep/, $testItems[0];
  $maxvar = scalar split /$smallsep/, $item;

}

############################################################
sub parseTestItem {

## parse test item

  my $item;
  ($curTestOutcome, $item, $curTestSpec)
    = split /$bigsep/, (shift @testItems), 3;
  my $eq = ($excNull ? $item =~ tr/=/=/ : 0);
  @curTestItem = split /$smallsep/, $item;
  $activeVar = @curTestItem - $eq;
  $activeContexts = 1 << $activeVar;
}

############################################################
sub beginTestItem {

## initialize count

  @datacount = @sum = ();
  $grandtotal = 0;

## select data items by probability

  $excludedData = 0;
  $probvec = pack "b*", "1" x @data;
  if ($probability < 1) {
    for (my $i = 0; $i < @data; ++$i) {
      vec($probvec, $i, 1) = 0, ++$excludedData
	if rand() > $probability;
    }
  }

## print header

  my ($sec,$min,$hour) = localtime();
  print "Time: ${hour}h ${min}m ${sec}s\n\n";
  print STDOUT "Time: ${hour}h ${min}m ${sec}s\n\n";
  print STDOUT "@{[@curTestItem]}\n";

## compare data and given

  undef %subcontext;
  undef %subtooutcome;
  my $testindata = 0;
  my $eg = 0;
  my $i;
  for ($i = 0; $i < @data; $i++) {
    next unless vec($probvec, $i, 1);
    my @dataItem = @{$data[$i]};
    my $context = 0;
    my $j;
    for ($j = 0; $j < @curTestItem; $j++) {
      next if $curTestItem[$j] eq '=';
      my $bit = $curTestItem[$j] ne $dataItem[$j];
      $context = ($context << 1) | $bit;
    }
    push @{$subcontext{$context}}, $i;
    my $outcome = $outcome[$i];
    next if ($subtooutcome{$context} == $outcome);
    $subtooutcome{$context} = $outcome, next
      unless defined $subtooutcome{$context};
    $subtooutcome{$context} = 0;
  }

  if (exists $subcontext{0}) {
    ++$testindata;
    delete $subcontext{0}, ++$eg if $excGiven;
  }
  my(@subcontexts) = keys %subcontext;
  $numsubcontexts = scalar @subcontexts;
  $subcontexts = pack "L*", @subcontexts;
  $suboutcomes = pack "C*", map { $subtooutcome{$_} } @subcontexts;

  print <<TOP;                     # here document
Project Name: $name
Given Context:  @curTestItem, $curTestSpec
@{[ $excGiven ? "If context is in data file then exclude"
    : "Include context even if it is in the data file" ]}
Number of data points: @{[ scalar @data ]}
Probability of including any one data point: $probability
Total Excluded: @{[ $eg ? "$excludedData + test item" : $excludedData ]}
Nulls: @{[ $excNull ? $excNull : "include" ]}
Gang: $linearOrSq
Number of active variables: $activeVar
Number of active contexts: $activeContexts
TOP
  print "Test item is in the data.\n" if $testindata;
  print "\n\n";
}

############################################################
sub count {
  my $count = $_[0];
  @list = ();
  foreach (unpack "L*", $_[1]) {
    push @list, @{$subcontext{$_}};
  }
  foreach (@list) {
    $datacount[$_] += (1 + $$tally) * $count;
    $sum[$outcome[$_]] += (1 + $$tally) * $count;
    $grandtotal += (1 + $$tally) * $count;
  }
}

############################################################
sub endTestItem {

  if ($grandtotal) {
    print "Analogical Set\n\n";
    print "Total Frequency = $grandtotal\n\n";
    my $i;
    for ($i = 0; $i < @datacount; $i++) {
      next unless $datacount[$i];
      printf "%-9.9s\t%-11.11s\t%7.0f\t%5.2f%%\n",
        $ocl[$outcome[$i]],
        $spec[$i],
        $datacount[$i],
        100 * $datacount[$i]/$grandtotal;
    }
    print "\n\n";
  }

  if (!$grandtotal) {
    print "No data items considered.  No prediction possible.\n\n\n";
  } else {
    print "Statistical Summary\n\n";
    my $high = 0;
    for ($i = 1; $i < @outcomelist; ++$i) {
      my $n;
      next unless $n = $sum[$i];
      $high = $n if $n > $high;
      printf "%-9.9s\t%-11.11s\t%7.0f\t%5.2f%%\n",
        $ocl[$i],
        $outcomelist[$i],
        $n,
        100 * $n/$grandtotal;
    }
    if ($sum[$curTestOutcome] == $high) {
      print "\nCorrect outcome predicted.\n";
      ++$totalCorrectPredictions;
    }
    print "\n\n";
  }
}

############################################################
sub closeshop {
  my ($sec,$min,$hour) = localtime();
  print "Time: ${hour}h ${min}m ${sec}s\n\n";
  print STDOUT "Time: ${hour}h ${min}m ${sec}s\n\n";

  if ($outputToFile eq 'yes') {
    close PFLE;
    print STDOUT "File is done.  Look at $resultsfile\n";
  }
}
